import requests
import tempfile
import platform
from pathlib import Path
from typing import Tuple
from datomizer import Datomizer
from datomizer.utils import constants, general


def create_origin_private_datasource_put_presigned(datomizer: Datomizer, path: str) -> Tuple[int, str]:
    datasource = datomizer.get_response_json(requests.post, url=constants.MANAGEMENT_POST_ADD_ORIGIN_PRIVATE_DATASOURCE)
    put_presigned_response = datomizer.get_response_json(requests.get,
                                                         url=constants.MANAGEMENT_GET_PUT_PRESIGNED_URL,
                                                         url_params=[datasource[general.ID], path])
    return datasource[general.ID], put_presigned_response[general.URL]


def create_origin_private_datasource_from_path(datomizer: Datomizer, path: Path):
    datasource_id, put_presigned_url = create_origin_private_datasource_put_presigned(datomizer, path.name)
    upload_temp_file(path, put_presigned_url)
    return datasource_id


def create_origin_private_datasource_from_df(datomizer: Datomizer, df, name="temp") -> int:
    if not name.endswith(".csv"):
        name = name + ".csv"

    datasource_id, put_presigned_url = create_origin_private_datasource_put_presigned(datomizer, name)
    upload_dataframe(put_presigned_url, df, name)
    return datasource_id


def create_target_private_datasource(datomizer: Datomizer) -> int:
    datasource = datomizer.get_response_json(requests.post, url=constants.MANAGEMENT_POST_ADD_TARGET_PRIVATE_DATASOURCE)
    return datasource[general.ID]


def upload_dataframe(presigned_url: str, df, name: str) -> None:
    with tempfile.TemporaryDirectory() as temp_dir:
        path = f"{temp_dir}/{name}"
        df_type = f"{type(df).__module__}.{type(df).__name__}"
        if df_type == "pandas.core.frame.DataFrame":
            df.to_csv(path, index=False)
        upload_temp_file(path, presigned_url)


def upload_temp_file(path, presigned_url):
    with open(path, 'rb') as temp_file:
        response = requests.put(url=presigned_url, data=temp_file)
        Datomizer.validate_response(response, "Upload to put presigned url")



import json
import requests
from datomizer import DatoMapper
from datomizer.utils import constants, general


def create_train_request(should_evaluate=True) -> dict:
    return {"sampleInputRatio": 1, "privacyLevels": [-1], "shouldEvaluate": should_evaluate}


def train(dato_mapper: DatoMapper, should_evaluate=True) -> int:
    dato_mapper.wait()
    response_json = dato_mapper.datomizer.get_response_json(requests.put,
                                                            url=constants.MANAGEMENT_PUT_TRAIN_FLOW,
                                                            url_params=[dato_mapper.business_unit_id,
                                                                        dato_mapper.project_id,
                                                                        dato_mapper.flow_id],
                                                            headers={"Content-Type": "application/json"},
                                                            data=json.dumps(create_train_request(should_evaluate)))
    return response_json


def get_train_iteration(dato_mapper: DatoMapper, train_id) -> dict:
    flow = dato_mapper.get_flow()
    iteration = [t for t in flow[general.TRAIN_ITERATIONS] if t[general.ID] == train_id][0]
    return iteration


def get_train_generated_zip(dato_mapper: DatoMapper, train_id: int) -> str:
    response_json = dato_mapper.datomizer.get_response_json(requests.get,
                                                            url=constants.MANAGEMENT_GET_GENERATED_DATA,
                                                            url_params=[dato_mapper.business_unit_id,
                                                                        dato_mapper.project_id,
                                                                        dato_mapper.flow_id, train_id])
    return response_json[general.URL]



ID = "id"
URL = "url"


ERROR = "ERROR"
INITIALIZING = "INITIALIZING"
IN_PROGRESS = "IN_PROGRESS"

AUTHORIZATION = "Authorization"
BEARER_TOKEN = "Bearer %s"
D_REALM = "d-realm"
ACCESS_TOKEN = "access_token"
REFRESH_TOKEN = "refresh_token"

TRAIN_ITERATIONS = "trainIterations"
STEPS = "steps"
ORIGIN_DATASOURCE_ID = "originDataSourceId"
TARGET_DATASOURCE_ID = "targetDataSourceId"
MODELS = "models"


NEW_PRIVATE_ID = -999
NEW_SAMPLE_ID = -888

SPARK_WRITE_OPTIONS_OVERWRITE = "overwrite"
SPARK_WRITE_OPTIONS_APPEND = "append"

ENHANCE_ML_TARGET_COLUMN = "targetColumnName"
ENHANCE_ML_TARGET_TABLE = "targetTableName"



import json
import time
import requests
from datomizer import Datomizer
from datomizer.utils import constants, general


def create_flow_creation_request(datasource_id: int, sample_percent: int = 1, title: str = "new_flow") -> dict:
    return {
        "modelTrainingConfiguration": {
            "title": title,
            "dataSourceId": datasource_id,
            "sampleInputData": sample_percent,
            "privacyLevels": [-1]
        }
    }


def discover(datomizer: Datomizer,
             business_unit_id: str, project_id: str, datasource_id: str,
             sample_percent: int = 1, title: str = "new_flow") -> int:
    flow_creation_request = create_flow_creation_request(datasource_id, sample_percent, title)
    response_json = datomizer.get_response_json(requests.post,
                                                url=constants.MANAGEMENT_POST_ADD_FLOW,
                                                url_params=[business_unit_id, project_id],
                                                headers={"Content-Type": "application/json"},
                                                data=json.dumps(flow_creation_request))
    return response_json[general.ID]


def get_schema_discovery(datomizer: Datomizer,
                         business_unit_id: str, project_id: str, flow_id: str) -> dict:
    response_json = datomizer.get_response_json(requests.get,
                                                url=constants.MANAGEMENT_GET_SCHEMA_DISCOVERY,
                                                url_params=[business_unit_id, project_id, flow_id])
    return response_json




URL = "https://%s"
DEFAULT_DOMAIN = "%s.datomize.com"

MANAGEMENT = "/api"
ONBOARDING = "/onboarding"
IDENTITY = "/identity"
KEYCLOAK = "/auth"

VERSION = "/v1"
TENANT = "/tenant"


TOKEN_RESOLVER = "/tokenResolver"
REALM_NAME = "/realmName"

KEYCLOAK_BASE_URL = URL + KEYCLOAK + "/realms/%s/protocol/openid-connect/"
KEYCLOAK_GET_TOKEN_URL = KEYCLOAK_BASE_URL + "token"
KEYCLOAK_LOGOUT_URL = KEYCLOAK_BASE_URL + "logout"


ONBOARDING_GET_USER_DOMAIN_DEFAULT = (URL % DEFAULT_DOMAIN) \
                                     + ONBOARDING + VERSION + TENANT + "/getDefaultTenantDomainByKey/%s"

IDENTITY_GET_REALM_BY_DOMAIN = URL + IDENTITY + TOKEN_RESOLVER + REALM_NAME

# Temp
FLOWS_FILTER = "/flowsFilter"
LATEST_FLOWS = "/latestFlows"
MANAGEMENT_GET_LATEST_FLOWS = URL + MANAGEMENT + FLOWS_FILTER + LATEST_FLOWS


DATASOURCE_URL = "/dataSourceConfigurations"
ORIGIN_DS = "/origin"
TARGET_DS = "/target"
PRIVATE = "/private"
PRESIGNED_URL = "/presignedUrl"
MANAGEMENT_POST_ADD_ORIGIN_PRIVATE_DATASOURCE = URL + MANAGEMENT + DATASOURCE_URL + ORIGIN_DS + PRIVATE
MANAGEMENT_POST_ADD_TARGET_PRIVATE_DATASOURCE = URL + MANAGEMENT + DATASOURCE_URL + TARGET_DS + PRIVATE
MANAGEMENT_GET_PUT_PRESIGNED_URL = URL + MANAGEMENT + DATASOURCE_URL + ORIGIN_DS + PRIVATE + "/%s" + PRESIGNED_URL + "/%s"

PROTO = "/proto"
BUSINESS_UNIT_URL = "/businessUnit"
PROJECT_URL = "/projects"
GENERATIVE_MODEL = "/generativeModelFlows"
SYNTH_FLOW = "/synthesizeFlows"
FLOWS_URL = "/flows"
AUTO_SCHEMA_DISCOVERY = "/autoSchemaDiscovery"
SAMPLE_DATA_CSV = "/sampleTableCSV"
GENERATED_DATA_CSV = "/generatedTableCSV"
TRAIN = "/train"
ENHANCE_ML = "/enhanceMLFlows"
ENHANCE_ML_EVALUATE = "/enhancedMLEvaluate"
ENHANCE_ML_TRAIN_GEN = "/enhancedMLTrainAndGenerate"
MODEL = "/models"
SAMPLE_TABLE = "/sampleTable"
GENERATED_DATA = "/generatedData"
SYNTHESIZE = "/synthesize"

SINGLE_PROJECT_URL = URL + MANAGEMENT + BUSINESS_UNIT_URL + "/%s" + PROJECT_URL + "/%s"

MANAGEMENT_GET_ALL_BUSINESS_UNIT_URL = URL + MANAGEMENT + BUSINESS_UNIT_URL
MANAGEMENT_GET_GENERATIVE_FLOW = SINGLE_PROJECT_URL + GENERATIVE_MODEL + "/%s"
MANAGEMENT_GET_SYNTH_FLOW = SINGLE_PROJECT_URL + SYNTH_FLOW + "/%s"
MANAGEMENT_POST_ADD_FLOW = SINGLE_PROJECT_URL + GENERATIVE_MODEL
MANAGEMENT_GET_SCHEMA_DISCOVERY = SINGLE_PROJECT_URL + FLOWS_URL + "/%s" + AUTO_SCHEMA_DISCOVERY
MANAGEMENT_PUT_TRAIN_FLOW = SINGLE_PROJECT_URL + GENERATIVE_MODEL + "/%s" + TRAIN
MANAGEMENT_PUT_ENHANCE_ML_EVALUATE_FLOW = SINGLE_PROJECT_URL + ENHANCE_ML + "/%s" + ENHANCE_ML_EVALUATE
MANAGEMENT_PUT_ENHANCE_ML_TRAIN_GEN_FLOW = SINGLE_PROJECT_URL + ENHANCE_ML + "/%s" + ENHANCE_ML_TRAIN_GEN
MANAGEMENT_GET_GENERATIVE_SAMPLE = SINGLE_PROJECT_URL + GENERATIVE_MODEL + "/%s" + MODEL + "/%s" + SAMPLE_TABLE + "/%s"
MANAGEMENT_GET_GENERATED_DATA = SINGLE_PROJECT_URL + SYNTH_FLOW + "/%s" + TRAIN + "/%s" + GENERATED_DATA
MANAGEMENT_GET_GENERATED_DATA_CSV = SINGLE_PROJECT_URL + SYNTH_FLOW + "/%s" + TRAIN + "/%s" + GENERATED_DATA_CSV + "/%s"
MANAGEMENT_GET_SAMPLE_CSV = SINGLE_PROJECT_URL + SYNTH_FLOW + "/%s" + SAMPLE_DATA_CSV + "/%s"
MANAGEMENT_PUT_SYNTH_FLOW = SINGLE_PROJECT_URL + SYNTH_FLOW + "/%s" + SYNTHESIZE



from datomizer import Datomizer
from datomizer.helpers.datasource import datasource_helper
from datomizer.helpers import common_helper
from datomizer.helpers.business_unit_project import business_unit_project_helper as bu_proj_helper
from datomizer.helpers.autodiscovery import autodiscovery_helper
from datomizer.helpers.wrapper.schema_wrapper import SchemaWrapper
from datomizer.utils import step_types, general


class DatoMapper(object):
    datomizer: Datomizer
    business_unit_id = 0
    project_id = 0
    datasource_id = 0
    flow_id = 0
    schema: SchemaWrapper = None

    def __init__(self, datomizer: Datomizer):
        """Create DatoMapper object for extracting the structure of the input data.
        Args:
            datomizer: the Datomizer authentication object."""
        datomizer.next_step_validation()
        self.datomizer = datomizer
        self.business_unit_id, self.project_id = bu_proj_helper.get_default_business_unit_project(self.datomizer)

    @classmethod
    def restore(cls, datomizer: Datomizer, flow_id):
        dato_mapper = cls(datomizer)
        dato_mapper.flow_id = flow_id
        dato_mapper.wait()
        return dato_mapper

    def get_flow(self) -> dict:
        self.restore_validation()
        return common_helper.get_flow(self.datomizer, self.business_unit_id, self.project_id, self.flow_id)

    def create_datasource_from_path(self, path) -> None:
        self.set_datasource_validation()
        self.datasource_id = datasource_helper.create_origin_private_datasource_from_path(self.datomizer, path)

    def create_datasource_from_df(self, df, name="temp_df") -> None:
        self.set_datasource_validation()
        self.datasource_id = datasource_helper.create_origin_private_datasource_from_df(self.datomizer, df, name)

    def create_datasource(self, path, df, name):
        if path is not None:
            self.create_datasource_from_path(path)
        elif df is not None:
            self.create_datasource_from_df(df, name)

    def discover(self, path=None, df=None, df_name=None, sample_percent: int = 1, title: str = "sdk_flow", wait=True) -> None:
        """Extract the structure of the input data.
        Args:
            path: full path for the input data
            df: a dataframe containing the input data
            df_name: the name of the input data
            sample_percent: the sampling ratio; 1 (100%) by default.
            title: the name of the created task in Datomize; "sdk_flow" by default.
            wait: use wait=False for asynchronous programming; True by default (awaits for the results)."""
        self.create_datasource(path, df, df_name if df_name else "temp_sdk")

        self.pre_run_validation()
        self.flow_id = autodiscovery_helper.discover(self.datomizer,
                                                     self.business_unit_id, self.project_id, self.datasource_id,
                                                     sample_percent, title)
        if wait:
            self.wait()

    def wait(self) -> None:
        """Wait until the discovery method returns."""
        self.restore_validation()
        status = common_helper.wait_for_step_type(datomizer=self.datomizer,
                                                  business_unit_id=self.business_unit_id,
                                                  project_id=self.project_id,
                                                  flow_id=self.flow_id,
                                                  step_type=step_types.COLUMN_DISCOVERY)
        if status == general.ERROR:
            raise Exception("Auto Discovery Failed")
        self.datasource_id = self.get_flow()[general.ORIGIN_DATASOURCE_ID]
        self.get_schema_discovery()

    def get_schema_discovery(self) -> dict:
        """Get a data mapping object for the input data.
        Returns:
            Datomize mapping dictionary"""
        self.restore_validation()
        if self.schema:
            return
        self.schema = SchemaWrapper(autodiscovery_helper
                                    .get_schema_discovery(self.datomizer,
                                                          self.business_unit_id, self.project_id, self.flow_id))
        return self.schema.schema

    def base_validation(self):
        if not (self.business_unit_id > 0 and self.project_id > 0):
            raise Exception("missing base properties")

    def set_datasource_validation(self):
        self.base_validation()
        if self.datasource_id > 0:
            raise Exception("datasource id cannot be mutated")

    def pre_run_validation(self):
        self.base_validation()
        if not (self.datasource_id > 0):
            raise Exception("datasource id required for this step")

    def restore_validation(self):
        self.base_validation()
        if not (self.flow_id > 0):
            raise Exception("flow id required for this step")

    def next_step_validation(self):
        self.restore_validation()
        if not self.schema:
            raise Exception("DatoMapper not ready")



import enum


class Algorithms(enum.Enum):
    ADABOOST = "ADABOOST"
    BGG = "BGG"
    CATBOOST = "CATBOOST"
    DECISION_TREE = "DECISION_TREE"
    EXTRA_TREE = "EXTRA_TREE"
    GBM = "GBM"
    KNN = "KNN"
    LINEAR_SVM = "LINEAR_SVM"
    LR = "LR"
    LGBM = "LGBM"
    NB = "NB"
    NN = "NN"
    POLY_SVM = "POLY_SVM"
    RANDOM_FOREST = "RANDOM_FOREST"
    SVM = "SVM"
    XGBOOST = "XGBOOST"


class Metrics(enum.Enum):
    ACC = "ACC"
    AUC = "AUC"
    BALANCED_ACC = "BALANCED_ACC"
    F1 = "F1"
    MAE = "MAE"
    MSE = "MSE"
    PRECISION = "PRECISION"
    RECALL = "RECALL"
    RMSE = "RMSE"
    ROC_AUC = "ROC_AUC"



import requests
from datomizer import Datomizer
from datomizer.utils import constants, general
from typing import Tuple


def get_default_business_unit_project(datomizer: Datomizer) -> Tuple[int, int]:
    business_units = datomizer.get_response_json(method=requests.get,
                                                 url=constants.MANAGEMENT_GET_ALL_BUSINESS_UNIT_URL,
                                                 params={"filterProjectAdminOrMember": True})
    return business_units[0][general.ID], business_units[0]['projects'][0][general.ID]



import requests
from datomizer.utils import constants

CLIENT = "direct-sdk"


def get_domain_by_username(username: str, env: str = 'app') -> str:
    response = requests.get(constants.ONBOARDING_GET_USER_DOMAIN_DEFAULT % (env, username))
    return response.url.replace("https://", "").split('/')[0]


def get_realm_by_domain(domain: str) -> str:
    response = requests.get(constants.IDENTITY_GET_REALM_BY_DOMAIN % domain)
    return response.text


def post_token(realm: str, domain: str, username: str, password: str):
    client_props = {
        "client_id": CLIENT,
        "grant_type": "password",
        "username": username,
        "password": password
    }

    response = requests.post(constants.KEYCLOAK_GET_TOKEN_URL % (domain, realm), client_props)
    return response.json()


def post_refresh_token(realm: str, domain: str, token: str):
    client_props = {
        "client_id": CLIENT,
        "grant_type": "refresh_token",
        "refresh_token": token
    }

    response = requests.post(constants.KEYCLOAK_GET_TOKEN_URL % (domain, realm), client_props)
    return response.json()


def post_log_out(realm: str, domain: str, token: str, refresh_token: str):
    headers = {"Authorization": f"Bearer {token}"}
    client_props = {
        "client_id": CLIENT,
        "refresh_token": refresh_token
    }
    response = requests.post(constants.KEYCLOAK_LOGOUT_URL % (domain, realm), client_props, headers=headers)
    return response.text



import io

from datomizer import DatoMapper
from datomizer.helpers import common_helper
from datomizer.helpers.train import enhance_ml_helper, train_helper
from datomizer.utils import general, step_types
from datomizer.helpers.wrapper import schema_wrapper
from datomizer.utils.enhance_ml import Metrics, Algorithms


class DatoEnhancer(object):
    dato_mapper: DatoMapper
    train_id = 0
    model_id = 0
    target_table = ""
    target_column = ""

    def __init__(self, dato_mapper: DatoMapper):
        """Create DatoTrainer object for training a generative model for the mapped input data.
        Args:
            dato_mapper: the DatoMapper object for the input data."""
        dato_mapper.next_step_validation()
        self.dato_mapper = dato_mapper

    @classmethod
    def restore(cls, dato_mapper: DatoMapper, train_id):
        dato_enhancer = cls(dato_mapper)
        dato_enhancer.train_id = train_id
        dato_enhancer.wait()
        return dato_enhancer

    def enhance(self, target_table: str = "", target_column: str = "",
                metric_list: [Metrics] = [], algorithm_list: [Algorithms] = [], wait=True) -> None:
        if self.train_id > 0:
            return

        metric_list = [metric.value for metric in metric_list]
        algorithm_list = [algorithm.value for algorithm in algorithm_list]

        self.target_table = self.dato_mapper.schema.table(target_table)[schema_wrapper.NAME]
        self.target_column = self.dato_mapper.schema.column(target_table, target_column)[schema_wrapper.NAME]

        self.train_id = enhance_ml_helper.enhance_ml_evaluate(self.dato_mapper,
                                                              target_table=self.target_table,
                                                              target_column=self.target_column,
                                                              metric_list=metric_list, algorithm_list=algorithm_list)
        if wait:
            self.wait()

    def wait(self) -> None:
        """Wait until the train method returns."""
        self.restore_validation()
        status = common_helper.wait_for_step_type(datomizer=self.dato_mapper.datomizer,
                                                  business_unit_id=self.dato_mapper.business_unit_id,
                                                  project_id=self.dato_mapper.project_id,
                                                  flow_id=self.dato_mapper.flow_id,
                                                  step_type=step_types.ML_TRAIN_AND_GENERATE,
                                                  train_id=self.train_id)
        if status == general.ERROR:
            raise Exception("Trainer Failed")
        train = train_helper.get_train_iteration(self.dato_mapper, self.train_id)
        self.model_id = train[general.MODELS][0][general.ID]

    def get_generated_data(self) -> None:
        self.restore_validation()
        print(common_helper.get_generated_zip(datomizer=self.dato_mapper.datomizer,
                                              business_unit_id=self.dato_mapper.business_unit_id,
                                              project_id=self.dato_mapper.project_id,
                                              flow_id=self.dato_mapper.flow_id,
                                              train_id=self.train_id))

    def get_generated_data_csv(self, table_name: str = None) -> io.StringIO:
        """Get the generated data in a csv format.
                Args:
                    table_name: the name of the generated data
                Returns:
                    StringIO object containing the generated data"""
        self.restore_validation()

        table_name = self.dato_mapper.schema.table(table_name)[schema_wrapper.NAME]

        return common_helper.get_generated_csv(datomizer=self.dato_mapper.datomizer,
                                               business_unit_id=self.dato_mapper.business_unit_id,
                                               project_id=self.dato_mapper.project_id,
                                               flow_id=self.dato_mapper.flow_id,
                                               train_id=self.train_id,
                                               table_name=table_name)

    def restore_validation(self):
        if not (self.train_id > 0):
            raise Exception("flow id required for this step")



import io

import requests
import time
from datomizer import Datomizer
from datomizer.utils import constants, step_types, general


def get_flow(datomizer: Datomizer,
             business_unit_id: str, project_id: str, flow_id: str, is_synth=False) -> dict:

    url = constants.MANAGEMENT_GET_SYNTH_FLOW if is_synth else constants.MANAGEMENT_GET_GENERATIVE_FLOW
    return datomizer.get_response_json(requests.get,
                                       url=url,
                                       url_params=[business_unit_id, project_id, flow_id])


def get_flow_step_status(datomizer: Datomizer,
                         business_unit_id: str, project_id: str, flow_id: str, is_synth=False,
                         step_type: str = '', train_id: int = 0) -> str:
    response_json = get_flow(datomizer, business_unit_id, project_id, flow_id, is_synth)
    return get_relevant_steps_status(response_json, step_type, train_id)


def get_relevant_steps_status(response_json: {}, step_type: str = '', train_id: int = 0) -> str:
    if train_id > 0:
        steps = [iteration for iteration in response_json[general.TRAIN_ITERATIONS]
                 if iteration[general.ID] == train_id][0][general.STEPS]
    else:
        steps = response_json[general.STEPS]

    return get_steps_status(steps, step_type)


def get_steps_status(steps: [], step_type) -> str:
    for step in steps:
        for task in step['tasks']:
            if task['taskStatus'] == general.ERROR:
                return general.ERROR
            if step["stepType"] == step_type:
                return task['taskStatus']

    return general.INITIALIZING


def wait_for_step_type(datomizer: Datomizer,
                       business_unit_id: str, project_id: str, flow_id: str, is_synth=False,
                       step_type: str = '', train_id: int = 0) -> str:
    status = get_flow_step_status(datomizer, business_unit_id, project_id, flow_id, is_synth, step_type, train_id)
    while status in [general.INITIALIZING, general.IN_PROGRESS]:
        time.sleep(10)
        status = get_flow_step_status(datomizer, business_unit_id, project_id, flow_id, is_synth, step_type, train_id)
    return status


def get_generated_zip(datomizer: Datomizer,
                      business_unit_id: int, project_id: int, flow_id: int,
                      train_id: int = 0) -> str:
    response_json = datomizer.get_response_json(requests.get,
                                                url=constants.MANAGEMENT_GET_GENERATED_DATA,
                                                url_params=[business_unit_id, project_id, flow_id, train_id])
    return response_json[general.URL]


def get_generated_csv(datomizer: Datomizer,
                      business_unit_id: int, project_id: int, flow_id: int, train_id: int = 0,
                      table_name: str = "") -> io.StringIO:
    response_json = datomizer.get_response_json(requests.get,
                                                url=constants.MANAGEMENT_GET_GENERATED_DATA_CSV,
                                                url_params=[business_unit_id, project_id,
                                                            flow_id, train_id, table_name])
    response = requests.get(response_json[general.URL])
    Datomizer.validate_response(response, "Unable to get CSV from put_presigned url")

    return io.StringIO(response.content.decode('utf-8'))



NAME = "name"
TABLES = "tables"
COLUMNS = "columns"


class SchemaWrapper(object):
    schema = {}

    def __init__(self, schema):
        self.schema = schema

    def tables(self) -> list:
        return self.schema[TABLES]

    def table(self, table_name):
        tables = self.tables()
        for table in tables:
            if table_name == table[NAME]:
                return table

        return tables[0]

    def columns(self, table_name) -> list:
        return self.table(table_name)[COLUMNS]

    def column(self, table_name, column_name) -> list:
        columns = self.columns(table_name)
        for column in columns:
            if column_name == column[NAME]:
                return column

        return columns[0]



from datomizer.datomizer import Datomizer
from datomizer.datomapper import DatoMapper
from datomizer.datotrainer import DatoTrainer
from datomizer.datogenerator import DatoGenerator
from datomizer.datoenhancer import DatoEnhancer



import json
import requests
from datomizer import DatoTrainer
from datomizer.utils import constants, general


def create_synth_request(dato_trainer: DatoTrainer,
                         datasource_id: int = general.NEW_PRIVATE_ID,
                         sample_output_ratio: float = 1) -> dict:
    return {
        "title": "Synth SDK",
        "datasourceId": datasource_id,
        "sizeInGB": 0,
        "sampleOutputRatio": sample_output_ratio,
        "modelId": dato_trainer.model_id,
        "synthStrategy": general.SPARK_WRITE_OPTIONS_OVERWRITE
    }


def generate(dato_trainer: DatoTrainer, datasource_id: int, output_ratio: float = 1) -> int:
    dato_trainer.wait()
    response_json = dato_trainer.dato_mapper.datomizer.get_response_json(
        requests.put,
        url=constants.MANAGEMENT_PUT_SYNTH_FLOW,
        url_params=[dato_trainer.dato_mapper.business_unit_id,
                    dato_trainer.dato_mapper.project_id,
                    dato_trainer.dato_mapper.flow_id],
        headers={"Content-Type": "application/json"},
        data=json.dumps(create_synth_request(dato_trainer, datasource_id, output_ratio)))
    return response_json[general.ID]



import requests
from datomizer.helpers.authentication import authentication_helper
from datomizer.utils.general import ACCESS_TOKEN, REFRESH_TOKEN, AUTHORIZATION, BEARER_TOKEN, D_REALM


class Datomizer(object):
    domain: str
    realm: str
    access_token: str
    refresh_token: str
    auth_headers: dict

    def __init__(self, username: str, password: str, env: str = 'app'):
        """Create an authentication object and register Datomize on it using the Datomizer constructor

        Args:
            username: your username
            password: your password
        Returns:
            Datomizer Authentication object"""
        self.domain = authentication_helper.get_domain_by_username(username=username, env=env)
        self.realm = authentication_helper.get_realm_by_domain(domain=self.domain)
        token_response = authentication_helper.post_token(realm=self.realm, domain=self.domain,
                                                          username=username, password=password)
        self.config_token(token_response)

    def refresh_token_headers(self):
        token_response = authentication_helper.post_refresh_token(realm=self.realm,
                                                                  domain=self.domain,
                                                                  token=self.refresh_token)
        self.config_token(token_response)

    def config_token(self, token_response):
        try:
            self.access_token = token_response[ACCESS_TOKEN]
            self.refresh_token = token_response[REFRESH_TOKEN]
            self.auth_headers = {AUTHORIZATION: BEARER_TOKEN % self.access_token, D_REALM: self.realm}
        except Exception as ex:
            print(f"token res:{token_response}")
            raise ex

    def get_response_json(self, method: requests.Request, url: str, headers: {} = {}, url_params: [] = [], **kwargs):
        return self.api_request(method=method, url=url, headers=headers, url_params=url_params, **kwargs).json()

    def api_request(self, method: requests.Request,
                    url: str, headers: {} = {}, url_params: [] = [], **kwargs) -> requests.Response:
        kwargs['url'] = url % (self.domain, *url_params)

        headers.update(self.auth_headers)
        kwargs['headers'] = headers
        response: requests.Response = method(**kwargs)

        # refresh token and retry on unauthorized or forbidden request
        if response.status_code in [401, 403]:
            self.refresh_token_headers()
            headers.update(self.auth_headers)
            response = method(**kwargs)

        self.validate_response(response)
        return response

    def base_validation(self):
        if not (self.realm and self.domain):
            raise Exception("missing base properties")

    def next_step_validation(self):
        if not self.access_token:
            raise Exception("access token is not configured")

    @staticmethod
    def validate_response(response: requests.Response, not_valid_message: str = "Status is not OK"):
        if response.status_code != 200:
            print(response.status_code)
            print(response.url)
            print(response.content)
            raise Exception(not_valid_message)

    def __del__(self):
        authentication_helper.post_log_out(realm=self.realm, domain=self.domain,
                                           token=self.access_token, refresh_token=self.refresh_token)



from setuptools import setup, find_namespace_packages

setup(
    name="datomize",
    version="0.0.1032",
    license='Apache Software License (https://www.apache.org/licenses/LICENSE-2.0)',
    description="Datomize python client",
    packages=find_namespace_packages(),
    install_requires=[
        'requests',
    ],
    url="https://datomize.github.io/datomizeSDK",
    author="Datomize Ltd.",
    author_email="support@datomize.com",
    project_urls={
        "Documentation": "https://datomize.github.io/datomizeSDK",
    },
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    keywords=['machine learning'],
    classifiers=[
          'Development Status :: 3 - Alpha',
          'License :: OSI Approved :: Apache Software License',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: 3.8',
          'Programming Language :: Python :: 3.9',
    ],
    python_requires='>=3.6',

)



import json
import requests
from datomizer import DatoMapper
from datomizer.utils import constants, general


def create_enhance_ml_request(target_table, target_column, metric_list=[], algorithm_list=[]) -> dict:
    return {"tableName": target_table, "targetColumn": target_column,
            "metricList": metric_list, "algorithmList": algorithm_list}


def enhance_ml_evaluate(dato_mapper: DatoMapper, target_table, target_column, metric_list=[], algorithm_list=[]) -> int:
    dato_mapper.wait()
    response_json = dato_mapper.datomizer.get_response_json(requests.put,
                                                            url=constants.MANAGEMENT_PUT_ENHANCE_ML_EVALUATE_FLOW,
                                                            url_params=[dato_mapper.business_unit_id,
                                                                        dato_mapper.project_id,
                                                                        dato_mapper.flow_id],
                                                            headers={"Content-Type": "application/json"},
                                                            data=json.dumps(create_enhance_ml_request(target_table,
                                                                                                      target_column,
                                                                                                      metric_list,
                                                                                                      algorithm_list)))
    return response_json


def get_train_iteration_by_target(dato_mapper: DatoMapper, target_table, target_column) -> dict:
    flow = dato_mapper.get_flow()
    iteration = [t for t in flow[general.TRAIN_ITERATIONS]
                 if t[general.ENHANCE_ML_TARGET_TABLE] == target_table
                 and t[general.ENHANCE_ML_TARGET_COLUMN == target_column]][0]
    return iteration




import io

from datomizer import DatoMapper
from datomizer.helpers import common_helper
from datomizer.helpers.train import train_helper
from datomizer.helpers.wrapper import schema_wrapper
from datomizer.utils import general, step_types


class DatoTrainer(object):
    dato_mapper: DatoMapper
    train_id = 0
    model_id = 0
    evaluate = True

    def __init__(self, dato_mapper: DatoMapper):
        """Create DatoTrainer object for training a generative model for the mapped input data.
        Args:
            dato_mapper: the DatoMapper object for the input data."""
        dato_mapper.next_step_validation()
        self.dato_mapper = dato_mapper

    @classmethod
    def restore(cls, dato_mapper: DatoMapper, train_id):
        dato_trainer = cls(dato_mapper)
        dato_trainer.train_id = train_id
        dato_trainer.wait()
        return dato_trainer

    def train(self, should_evaluate=True, wait=True) -> None:
        """Train a generative model.
        Args:
            should_evaluate: evaluate the generated data; Evaluation is performed (True) by default.
            wait: use wait=False for asynchronous programming; True by default (awaits for the results)."""
        if self.train_id > 0:
            return
        self.train_id = train_helper.train(self.dato_mapper, should_evaluate)
        self.evaluate = should_evaluate
        if wait:
            self.wait()

    def wait(self) -> None:
        """Wait until the train method returns."""
        self.restore_validation()
        status = common_helper.wait_for_step_type(datomizer=self.dato_mapper.datomizer,
                                                  business_unit_id=self.dato_mapper.business_unit_id,
                                                  project_id=self.dato_mapper.project_id,
                                                  flow_id=self.dato_mapper.flow_id,
                                                  step_type=step_types.EVALUATE if self.evaluate else step_types.TRAIN,
                                                  train_id=self.train_id)
        if status == general.ERROR:
            raise Exception("Trainer Failed")
        self.model_id = train_helper.get_train_iteration(self.dato_mapper, self.train_id)[general.MODELS][0][general.ID]

    def get_generated_data(self) -> None:
        self.next_step_validation()
        print(common_helper.get_generated_zip(datomizer=self.dato_mapper.datomizer,
                                              business_unit_id=self.dato_mapper.business_unit_id,
                                              project_id=self.dato_mapper.project_id,
                                              flow_id=self.dato_mapper.flow_id,
                                              train_id=self.train_id))

    def get_generated_data_csv(self, table_name: str = None) -> io.StringIO:
        """Get the generated data in a csv format.
                Args:
                    table_name: the name of the generated data
                Returns:
                    StringIO object containing the generated data"""
        self.next_step_validation()
        if not self.evaluate:
            return

        table_name = self.dato_mapper.schema.table(table_name)[schema_wrapper.NAME]

        return common_helper.get_generated_csv(datomizer=self.dato_mapper.datomizer,
                                               business_unit_id=self.dato_mapper.business_unit_id,
                                               project_id=self.dato_mapper.project_id,
                                               flow_id=self.dato_mapper.flow_id,
                                               train_id=self.train_id,
                                               table_name=table_name)

    def restore_validation(self):
        if not (self.train_id > 0):
            raise Exception("flow id required for this step")

    def next_step_validation(self):
        self.restore_validation()
        if not (self.model_id > 0):
            raise Exception("DatoTrainer not ready")



import io

from datomizer import DatoTrainer
from datomizer.helpers import common_helper
from datomizer.helpers.generator import generator_helper
from datomizer.helpers.datasource import datasource_helper
from datomizer.utils import general, step_types
from datomizer.helpers.wrapper import schema_wrapper


class DatoGenerator(object):
    dato_trainer: DatoTrainer
    synth_id = 0
    datasource_id = 0

    def __init__(self, dato_trainer: DatoTrainer):
        """Create DatoGenerator object for generating data using the trained generative model.
        Args:
            dato_trainer: the DatoTrainer object trained on the input data."""
        dato_trainer.next_step_validation()
        self.dato_trainer = dato_trainer

    @classmethod
    def restore(cls, dato_trainer: DatoTrainer, synth_id):
        dato_generator = cls(dato_trainer)
        dato_generator.synth_id = synth_id
        dato_generator.wait()
        return dato_generator

    def get_flow(self) -> dict:
        self.restore_validation()
        return common_helper.get_flow(datomizer=self.dato_trainer.dato_mapper.datomizer,
                                      business_unit_id=self.dato_trainer.dato_mapper.business_unit_id,
                                      project_id=self.dato_trainer.dato_mapper.project_id,
                                      flow_id=self.synth_id,
                                      is_synth=True)

    def create_datasource(self) -> None:
        if self.datasource_id > 0:
            return
        self.datasource_id = datasource_helper.create_target_private_datasource(self.dato_trainer.dato_mapper.datomizer)

    def generate(self, output_ratio: float = 1, wait=True) -> None:
        """Generate output data.
        Args:
            output_ratio: float represents the output ratio for generated data
            wait: use wait=False for asynchronous programming; True by default (awaits for the results)."""
        self.create_datasource()
        if self.synth_id > 0:
            return
        self.synth_id = generator_helper.generate(self.dato_trainer, self.datasource_id, output_ratio)
        if wait:
            self.wait()

    def wait(self) -> None:
        """Wait until the generate method returns."""
        self.restore_validation()
        status = common_helper.wait_for_step_type(datomizer=self.dato_trainer.dato_mapper.datomizer,
                                                  business_unit_id=self.dato_trainer.dato_mapper.business_unit_id,
                                                  project_id=self.dato_trainer.dato_mapper.project_id,
                                                  flow_id=self.synth_id,
                                                  is_synth=True,
                                                  step_type=step_types.GENERATE)
        if status == general.ERROR:
            raise Exception("Synth Failed")
        self.datasource_id = self.get_flow()[general.TARGET_DATASOURCE_ID]

    def get_generated_data(self) -> None:
        self.restore_validation()
        print(common_helper.get_generated_zip(datomizer=self.dato_trainer.dato_mapper.datomizer,
                                              business_unit_id=self.dato_trainer.dato_mapper.business_unit_id,
                                              project_id=self.dato_trainer.dato_mapper.project_id,
                                              flow_id=self.synth_id))

    def get_generated_data_csv(self, table_name: str = None) -> io.StringIO:
        """Get the generated data in a csv format.
                Args:
                    table_name: the name of the generated data
                Returns:
                    StringIO object containing the generated data"""
        self.restore_validation()

        table_name = self.dato_trainer.dato_mapper.schema.table(table_name)[schema_wrapper.NAME]

        return common_helper.get_generated_csv(datomizer=self.dato_trainer.dato_mapper.datomizer,
                                               business_unit_id=self.dato_trainer.dato_mapper.business_unit_id,
                                               project_id=self.dato_trainer.dato_mapper.project_id,
                                               flow_id=self.synth_id,
                                               table_name=table_name)

    def restore_validation(self):
        if not (self.synth_id > 0):
            raise Exception("synth id required for this step")
